#####################################################################
;# NAME:   dataGraph.tcl
;# AUTHOR: Clif Flynt
;# DATE:   10/2/2005
;# DESC:   A generic data manipulation and graphing
;#         application.
;#         
;#
;# PARAMETERS:  
;#  GUI
;#
#####################################################################
# This code is written and copyrighted by Clif Flynt, 2003
#
# It is licenced for public use with these terms:
#
#  This licence agreement must remain intact with the code.
# 
#  You may use this code freely for non-commercial purposes.
#
#  You may examine and modify the code to your heart's content.
#
#  You may not claim to have authored the code (aside from modifications
#  you may have made.)
#
#  You may not distribute the code without this license agreement
#
#  Maintenance, extension, modification, etc can be performed by:
#
#   Noumena Corporation
#   9300 Fleming Rd.
#   Dexter, MI  48130
#
#   Contact: clif@noucorp.com
#
#####################################################################

set release(dataGraph.tcl) {$Header: /home/cvs/dataGraph/dataGraph.tcl,v 1.3 2005/10/02 14:47:21 clif Exp $}
package require Tk

if {![namespace exists ::zvfs]} {
  set SRCDIR .
}
if {[namespace exists ::mk4vfs]} {
  set SRCDIR [file join $starkit::topdir lib app-dataGraph]
}

source $SRCDIR/graph2.tcl
source $SRCDIR/combobox.tcl
source $SRCDIR/notebook.tcl
source $SRCDIR/help.tcl
source $SRCDIR/htmllib.tcl
source $SRCDIR/imgs.tcl
source $SRCDIR/options.tcl

array set State {
  workingDir .
}
# Initialize graph States to default values.
graph2::initStateDefaults graphDefault

# Add data defaults
array set graphDefault {
  1,point {}
  2,point {}
  3,point {}
  4,point {}
  5,point {}
  6,point {}
  7,point {}
  xlegend {}
  ylegend {}
  zlegend {}
  0,axis,y,color gray50
  0,axis,x,color gray50
}


# Some graph defaults 
array set canvasDefault {
  background white
  borderwidth 2
  height 400
  width 500
  relief ridge
}

proc labelFormat {axis val} {
  global dataDef
  set val [expr $val $dataDef($axis,axis,Scale)]
  switch $dataDef($axis,axis,Style) {
    numeric {
      return [format $dataDef($axis,axis,Format) $val]
    } 
    date {
      return [clock format [expr int ($val)] -format $dataDef($axis,axis,Format)]
    }
  }
}

proc saveState {} {
    global graphState
    global canvasState
    global dataDef
    global State
    
    set ofN [tk_getSaveFile \
        -initialdir $State(workingDir) \
	-defaultextension .def \
        -filetypes {{{Graph Definition} .def} {{All Files} *}}]
    
    if {[string equal {} $ofN]} {return}

    set State(workingDir) [file dirname $ofN]
    set of [open $ofN w]
    foreach var {graphState canvasState dataDef} {
	puts $of [list array set $var [array get $var]]
    }
    close $of
}

proc loadState {} {
    global graphState
    global canvasState
    global dataDef
    global State

    set ifN [tk_getOpenFile \
        -initialdir $State(workingDir) \
	-defaultextension .def \
        -filetypes {{{Graph Definition} .def} {{All Files} *}}]

    if {[string equal {} $ifN]} {return}
    if {![file exists $ifN]} {
      tk_messageBox -type ok -message "$ifN does not exist"
      return
    }
    set State(workingDir) [file dirname $ifN]

    source $ifN
}

proc GUI {} {
    global graphState
    global dataDef
    # style: histogram, line, scatter, pie, area, colorbar
    # Specifics: histShowAbsolute, histStyle, rotate, drawPoint

    # Axis labels, format, numeric/textlist
    # Axis titles 
    # Axis margins: left,top,right,bottom 
    # Colors: axis, graph background
    # Fonts: axis, titles, labels, legend
    # Legend: title, use, relx, rely, anchor

    # dataset 1-9: input file, X filter, Y filter, Z filter
    #              field separator, use field?
    #              Scaling - multiplier, to fill
    #              Point
    #              Legend Title


    # FILE menu -load/save all settings

    set col 0
    set row 0

    set bf [frame .bf]
    grid $bf -row [incr row] -column $col -sticky ewn

    foreach bn {File Go Help} \
      side {left left right} \
      cmds {{Save saveState Load loadState Exit exit} \
          {showGraph} \
	  {About help::about Help help::help Versions help::versions}} {
      if {[llength $cmds] > 1} {
        set w [menubutton $bf.b_$bn -text $bn -menu $bf.b_$bn.m]
        pack $w -side $side
        menu $bf.b_$bn.m
        foreach {id cmd} $cmds {
          $bf.b_$bn.m add command -label $id -command $cmd
        }
      } else {
        set w [button $bf.b_$bn -text $bn -command $cmds]
        pack $w -side $side
      }
    }

    foreach in [array names help::helps] {
      if {[string first , $in] < 0} {continue}
      foreach {type item} [split $in ,] {break;}
      set title [string totitle $type]
      if {![info exists menus($title)]} {
        $bf.b_Help.m add cascade -label [string map {_ { }} $title] \
	    -menu $bf.b_Help.m.m_$title
        set menus($title) [menu $bf.b_Help.m.m_$title]
      }
        $bf.b_Help.m.m_$title add command -label [string map {_ { }} $item] \
	    -command "help::help $in"
    }

    set gf [labelframe .gd -text "Graph Description"]

    grid $gf -row [incr row] -column $col -sticky n

    set col 0
    set row -1

    ::notebook::create $gf.n -width 400 -height 450 -padx 5 -pady 5 \
        -pages \
      {Generic Line/Scatter Histogram Axis}
    grid $gf.n -row 0 -column 0 -sticky new
    # set f [labelframe $gf.l_Style -text Style]
    set f $gf.n.p0
#    grid $f -row [incr row] -column $col -sticky n -padx 2
    set ff [labelframe $f.f_Style -text "Graph Style"]
    grid $ff -row [incr row] -column $col  -columnspan 2 -sticky new
    set vals {histogram line scatter pie area colorbar}
    set lr 0; set lc 0
    foreach v $vals {
	set w [radiobutton $ff.b_$v -text $v -variable graphState(style)\
	  -value $v]
	grid $w -row $lr -column [incr lc] -sticky w
	if {$lc > 2} {
	  set lc 0; incr lr
	}
    }
    
    set graphState(style) line

    # set f [labelframe $gf.l_Settings -text "Generic Settings"]
    # grid $f -row [incr row] -column $col -sticky new -pady 5 -padx 2
#    set f $gf.n.p1
    foreach v {rotate } lbl {Rotation} {
        set ff [labelframe $f.l_Settings -text $lbl]
        grid $ff -row [incr row] -column $col -sticky new -pady 5 -padx 2 -columnspan 2
	set w [checkbutton $ff.b_$v -text $v -variable graphState($v) -onvalue 1]
	grid $w -row [incr lr] -column $lc -sticky w
	set graphState($v) 0
    }

    set vals {axis,x,label "X axis label"  {}
              axis,y,label "Y axis label"  {}}

    incr lr 
    foreach {v lbl dflt} $vals {
	set w [labelframe $f.f_$v -text $lbl]
	grid $w -row [incr lr] -column $lc

	set w [entry $w.e_$v -textvar graphState($v) -width 35]
	pack $w -side left -anchor w -padx 3
	set graphState($v) $dflt
    }

   # set f [labelframe $gf.l_LineSettings -text "Line/Scatter Settings"]
   # grid $f -row [incr row] -column $col -sticky new -pady 5 -padx 2
    set f $gf.n.p1
    set lr 0; set lc 0

    incr lr; 

    for {set ds 1} {$ds < 10} {incr ds} {
      set ff [labelframe $f.f_$ds -text "Dataset $ds"]
      grid $ff -row [incr lr] -column 0 -sticky w
      set lc -1

      foreach v {drawPoint} {
	set w [checkbutton $ff.b_$v -text $v \
	    -variable dataDef($ds,$v) -onvalue 1]
	grid $w -row $lr -column [incr lc] -sticky w
	set dataDef($ds,$v) 0
      }

      foreach v {dash width} dflt {- 1} {
        set fff [labelframe $ff.f_$v -text $v]
        grid $fff -row $lr -column [incr lc] -sticky w
	set w [entry $fff.e_$v -textvariable dataDef($ds,$v) -width 4]
	set dataDef($ds,$v) $dflt
	pack $w -side left
      }
      
      foreach v {color point} \
              dflt {black x} \
              vals {{red orange yellow green blue violet brown black white}
	            {+ x o box filled-o filled-box dot}} {
        set fff [labelframe $ff.f_$v -text $v]
        grid $fff -row $lr -column [incr lc] -sticky w
        ::combobox::create $fff.c_$v -width 12 -height 3 \
          -choices $vals -variable dataDef($ds,$v)
	set dataDef($ds,$v) $dflt
        pack $fff.c_$v  -side left
      }
      
	
      
    }

    set f $gf.n.p2
    # set f [labelframe $gf.l_HistSettings -text "Histogram Settings"]
    # grid $f -row [incr row] -column $col -sticky new -pady 5 -padx 2
    
    set lr 0; set lc -1

    foreach v {histShowAbsolute} ft {{Show Values}} {
      set f1 [labelframe $f.f_$v -text $ft]
      grid $f1 -row $lr -column [incr lc] 
      set w [checkbutton $f1.b_$v -text $v -variable graphState($v) -onvalue 1]
      grid $w
      set graphState($v) 0
    }
    
    set f1 [labelframe $f.st -text "Histogram Style"]
    grid $f1 -row $lr -column [incr lc] 
    ::combobox::create $f1.rules -width 10 -height 3 \
      -choices {stacked single} -variable graphState(histStyle)
    grid $f1.rules
    
    set col -1
    incr lr
    for {set ds 1} {$ds < 10} {incr ds} {
      set ff [labelframe $f.f_$ds -text "Dataset $ds"]
      grid $ff -row $lr -column [incr col] -sticky w  -padx 3 -pady 4
      if {$col > 0} {
        set col -1;
	incr lr
      }
      set lc -1

      foreach v {width} dflt {1} {
        set fff [labelframe $ff.f_$v -text $v]
        grid $fff -row $lr -column [incr lc] -sticky w
	set w [entry $fff.e_$v -textvariable dataDef($ds,$v) -width 4]
	set dataDef($ds,$v) $dflt
	pack $w -side left
      }
      
      foreach v {color } \
              dflt {black} \
              vals {{red orange yellow green blue violet brown black white}} {
        set fff [labelframe $ff.f_$v -text $v]
        grid $fff -row $lr -column [incr lc] -sticky w
        ::combobox::create $fff.c_$v -width 12 -height 3 \
          -choices $vals -variable dataDef($ds,$v)
	set dataDef($ds,$v) $dflt
        pack $fff.c_$v  -side left
      }
    }
    

    set f $gf.n.p3
    # set f [labelframe $gf.l_AxisSettings -text "Axis Settings"]
    # grid $f -row [incr row] -column $col -sticky new -pady 5 -padx 2

    set colors {red orange yellow green blue violet brown black white}

    set lr 0 
    set vals {0,axis,x,color "X axis color"  black
              0,axis,y,color "Y axis color"  black
              0,axis,x,lColor "X axis label color" black
              0,axis,y,lColor "Y axis label color" black
    }

    set lc 0
    foreach {v lbl dflt} $vals {
	set fff [labelframe $f.f_$v -text $lbl]
	grid $fff -row $lr -column [incr lc] -sticky ew -padx 3 -pady 2
	if {$lc > 1} {
	  set lc 0; incr lr
	}

        ::combobox::create $fff.c_$v -width 12 -height 3 \
          -choices $colors -variable graphState($v)
	set graphState($v) $dflt
        pack $fff.c_$v  -side left
    }

    set vals {x,axis,Style "X axis Format"  numeric
              y,axis,Style "Y axis Format"  numeric
    }

    set lc 0
    foreach {v lbl dflt} $vals {
	set fff [labelframe $f.f_$v -text $lbl]
	grid $fff -row $lr -column [incr lc] -sticky ew -padx 3 -pady 2
	if {$lc > 1} {
	  set lc 0; incr lr
	}

        ::combobox::create $fff.c_$v -width 12 -height 3 \
          -choices {numeric date} -variable dataDef($v)
	set dataDef($v) $dflt
        pack $fff.c_$v  -side left
    }

    set vals {x,axis,Format "X format String" "%.2f"
              y,axis,Format "Y format String" "%.2f"
              x,axis,Scale "X Scaler String" "* 1.0"
              y,axis,Scale "Y Scaler String" "* 1.0"
    }
    
    set lc 0
    foreach {v lbl dflt} $vals {
	set w [labelframe $f.f_$v -text $lbl]
	grid $w -row $lr -column [incr lc] -sticky ew -padx 3 -pady 2
	if {$lc > 1} {
	  set lc 0; incr lr
	}

	set w [entry $w.e_$v -textvar dataDef($v) -width 15]
	pack $w -side left -anchor w -padx 3
	set dataDef($v) $dflt
    }

    set vals {0,axis,x,LeftMargin "X axis Left Margin" 50
              0,axis,y,TopMargin "Y axis Top Margin"  25
              0,axis,x,RightMargin "X axis Right Margin" 20
              0,axis,y,BottomMargin "Y axis Bottom Margin" 35
    }
    
    set lc 0
    foreach {v lbl dflt} $vals {
	set w [labelframe $f.f_$v -text $lbl]
	grid $w -row $lr -column [incr lc] -sticky ew -padx 3 -pady 2
	if {$lc > 1} {
	  set lc 0; incr lr
	}

	set w [entry $w.e_$v -textvar graphState($v) -width 5]
	pack $w -side left -anchor w -padx 3
	set graphState($v) $dflt
    }

    set col 0

    set df [labelframe .dd -text "Data Definitions"]
    grid $df -row 1 -column [incr col] -rowspan 2 -sticky news
    grid rowconfigure . 1 -weight 1
    grid columnconfigure . $col -weight 1

    ::notebook::create $df.n -width 400 -height 450 -padx 5 -pady 5 -pages\
      {1 2 3 4 5 6 7 8 9}
    grid $df.n -sticky news -row 0 -column 0
    grid rowconfigure $df 0 -weight 1
    grid columnconfigure $df 0 -weight 1

    for {set i 1; set j 0} {$i < 10} {incr i; incr j} {
	makeDataDefGUI $df.n.p$j $i
    }
}

proc makeDataDefGUI {f id} {
    global dataDef
    global graphState
    set row 0
    set col 0

    set f1 [labelframe $f.lt -text "Legend Title"]
    grid $f1 -row [incr row] -column $col -sticky w -columnspan 3
    set w [entry $f1.lt -textvariable dataDef($id,legend) -width 30]
    grid $w
    
    incr row; set col -1

     set vals {"Use X Axis" axis,x,use "Use Y axis" axis,y,use}
     foreach {lbl v} $vals {
 	set w [checkbutton $f.b_$v -text $lbl -variable dataDef($id,$v) ]
        grid $w -row $row -column [incr col] -sticky w
 	set dataDef($id,$v) 1
     }

    set col 0
    set f1 [labelframe $f.x -text "X Values"]
    grid $f1 -row [incr row] -column $col -columnspan 3
    makeAxisEntryGUI $f1 $id x

    set f1 [labelframe $f.y -text "Y Values"]
    grid $f1 -row [incr row] -column $col -columnspan 3
    makeAxisEntryGUI $f1 $id y

    set f1 [labelframe $f.z -text "Z Values"]
    grid $f1 -row [incr row] -column $col -columnspan 3
    makeAxisEntryGUI $f1 $id z
}

proc makeAxisEntryGUI {f1 id axis} {
    global dataDef

    set row 0
    set col 0

    set ff [labelframe $f1.if -text "Data Source"]
    grid $ff -row [incr row] -column $col -columnspan 4
    set w [entry $ff.if -textvariable dataDef($id,$axis,if) -width 30]
    pack $w -side left
    set w [button $ff.br -text "Browse" -command "set dataDef($id,$axis,if)\
      \[tk_getOpenFile]"]
    pack $w -side right

    set ff [labelframe $f1.ptrn -text "Line Match Pattern"]
    grid $ff -row [incr row] -column 0 -columnspan 3
    set w [entry $ff.ptrn -textvariable dataDef($id,$axis,ptrn) -width 30]
    grid $w -sticky ew
    set dataDef($id,$axis,ptrn) *

    set ff [labelframe $f1.process -text "Process"]
    grid $ff -row $row -column 3
    ::combobox::create $ff.c_process -width 6 -height 3 \
        -choices {actual count index} -variable dataDef($id,$axis,process)
    set dataDef($id,$axis,process) actual
    grid $ff.c_process

    set ff [labelframe $f1.fld -text "Field"]
    grid $ff -row [incr row] -column $col
    set w [entry $ff.fld -textvariable dataDef($id,$axis,fld) -width 3]
    grid $w -sticky ew
    set dataDef($id,$axis,fld) 1

    set ff [labelframe $f1.sep -text "Separator"]
    grid $ff -row $row -column [incr col]
    set w [entry $ff.sep -textvariable dataDef($id,$axis,sep) -width 4]
    grid $w -sticky ew
    set dataDef($id,$axis,sep) ,

    set ff [labelframe $f1.format -text "Format"]
    grid $ff -row $row -column [incr col]
    ::combobox::create $ff.c_format -width 8 -height 3 \
          -choices {time date numeric {}} -variable dataDef($id,$axis,format)
    grid $ff.c_format
    set dataDef($id,$axis,format) {}

    set ff [labelframe $f1.scale -text "Scale By"]
    grid $ff -row $row -column [incr col]
    set w [entry $ff.scale -textvariable dataDef($id,$axis,scale) -width 6]
    grid $w -sticky ew
    set dataDef($id,$axis,scale) "* 1"
}

proc fillGraph {cvs dset xs ys zs title} {
    global graphState
    global dataDef
    
    set opts [list -color [lindex $graphState(colors) $dset] -legend,title $title]

    # Yanked from graph2::insertData
    foreach i {axis,x,use axis,y,use color point drawPoint selected dash \
               width tagList} {
      if {[info exists dataDef($dset,$i)]} {
	  lappend opts -$i [list $dataDef($dset,$i)]
      }
    }

    if {![string equal {} $zs]} {
        lappend opts -z $zs
    }
# puts "OPTS: $opts"
    eval graph2::insertData $cvs [list $xs] [list $ys] -dataset $dset $opts
}


foreach nm [array names graphDefault] {
    set graphState($nm) {}
}

foreach nm [array names canvasDefault] {
    set canvasState($nm) {}
}

proc loadData {id axis fileName} {
    global graphState
    global dataDef

    set hit 0

    if {[string first .gz $fileName] > 0} {
	set d [exec zcat $fileName]
    } {
	set if [open $fileName r]
	set d [read $if]
	close $if
    }
    foreach l [split $d \n] {
	# puts "[string match "*$dataDef($id,$axis,ptrn)*" $l] $l"

	if {[string match "*$dataDef($id,$axis,ptrn)*" $l]} {
	    set val [string trim [lindex [split $l\
	      $dataDef($id,$axis,sep)] $dataDef($id,$axis,fld)]]
	    if {[string equal {} $val]} {continue}
	    switch $dataDef($id,$axis,format) {
	      date -
	      time {
	        set val [clock scan $val]
	      }
	      numeric -
	      default {
	      }
	    }
	    # Scale it if numeric
	    if {[string is double $val]} {
	        set val [expr "$val $dataDef($id,$axis,scale)"]
	    }
	    # If process is count, increment a counter, else
	    #   grab the scaled data
	    switch $dataDef($id,$axis,process) {
	      actual {
	        lappend graphState($id,$axis) $val
	      }
	      index -
	      count {if {[info exists tmp($val)]} {
		        incr tmp($val)
		     } else {
		        set tmp($val) 1
		     }
	      }
	    }
	    set hit 1
	}
    }
    # If index, or count, collect data from tmp
    # and use that to fill the graphState variable
    switch $dataDef($id,$axis,process) {
      index {
        foreach i [lsort -dict [array names tmp]] {
	  lappend graphState($id,$axis) $i
	}
      }
      count {
        foreach i [lsort -dict [array names tmp]] {
	  lappend graphState($id,$axis) $tmp($i)
	}
      }
    }

    #  puts "RETURNING: $hit"
    return $hit
}

proc showGraph {} {
    global graphState
    global canvasState
    global graphDefault
    global canvasDefault
    global dataDef

    set gotData 0

    foreach nm [lsort -dict [array names dataDef *,y,if]] {
	foreach {id y if} [split $nm ,] {
	    break
	}

	foreach axis {x y z} {
	    set graphState($id,$axis) {}
	    if {![string equal {} $dataDef($id,$axis,if)]} {
		if {[loadData $id $axis $dataDef($id,$axis,if)]} {
		    set gotData 1
		}
	    }
	}
    }

    if {!$gotData} {
	tk_messageBox -type ok -message "Select some input data files"
	return
    }
    catch {destroy .t}
    catch {graph2::clearGraph .t.cr}
    toplevel .t

    foreach nm [array names canvasDefault] {
	if {[string equal {} $canvasState($nm)]} {
	    set canvasState($nm) $canvasDefault($nm)
	}
    }

    set options {}

    foreach arg [array names canvasState] {
	if {![string equal {} $canvasState($arg)]} {
	    lappend options -$arg $canvasState($arg)
	}
    }

    set cvs [eval canvas .t.cr $options]
    grid $cvs -row 1 -column 1 -sticky news

    grid rowconfigure .t 1 -weight 1
    grid columnconfigure .t 1 -weight 1

    foreach nm [array names graphDefault] {
	if {[string equal {} $graphState($nm)]} {
	    set graphState($nm) $graphDefault($nm)
	}
    }

    set options {-axis,x,nameProc {labelFormat x} -axis,y,nameProc {labelFormat y}}

    foreach arg [array names graphState] {
	if {[string match {[1-9],*} $arg]} {
	    continue
	}
	if {![string equal {} $graphState($arg)]} {
	    lappend options -$arg $graphState($arg)
	}
    }

    # puts "OPTIONS: $options"
    eval graph2::create $cvs $options

    grid rowconfigure . 1 -weight 1
    grid columnconfigure . 1 -weight 1

    foreach nm [array names graphState *,y] {
	foreach {dset axis} [split $nm ,] {
	    break
	}
	if {![string is double $dset]} {
	    continue
	}
	if {![string equal {} $graphState($nm)]} {
	    fillGraph $cvs $dset $graphState($dset,x) $graphState($dset,y)\
	      $graphState($dset,z) $dataDef($dset,legend)
	}

    }
    graph2::drawGraph $cvs
}

################################################################
# proc focusWindow {w}--
#    Raise, deiconfiy and place focus on window
# Arguments
#   w 		Name of window
# 
# Results
#   
# 
proc focusWindow {w} {
  set top [winfo toplevel $w]
  if {$top!=$w || [set target [::focus -lastfor $top]]==""} {
    set target $w
  }
  catch {::focus $target}
  wm deiconify $top
  # ::update
  raise $top
  ::focus $target
  after idle [subst -nocommands {catch {
    raise $top
    after idle {catch {::focus $target}}
  }}]
}

################################################################
# proc centerWindow {w {parent {}}}--
#    Put a window in the center of the screen or parent
# Arguments
#   w		Name of window
#   parent	Name of window to center in, default is screen.
# Results
#   
# 
proc centerWindow {w {parent {}}} {
  ::update
  set ww [winfo reqwidth $w]
  set wh [winfo reqheight $w]
  if {$parent==""} {
    set px 0
    set py 0
    set pw [winfo screenwidth .]
    set ph [winfo screenheight .]
    if {$pw>$ph*1.25} {set pw [expr {int($ph*1.25)}]}
    if {$ph>$pw*1.25} {set ph [expr {int($pw*1.25)}]}
  } else {
    set px [winfo rootx $parent]
    set py [winfo rooty $parent]
    set pw [winfo width $parent]
    set ph [winfo height $parent]
  }
  set x [expr {$px+($pw-$ww)/2}]
  set y [expr {$py+($ph-$wh)/2}]
  if {$x<0} {set x 0}
  if {$y<0} {set y 0}
  wm geometry $w +$x+$y
  focusWindow $w
}


GUI


